package CVSMonitor::Config;

# This is a wrapper over Config::Tiny for holding configuration
# information.

use strict;
use UNIVERSAL qw{isa can};
use base 'Config::Tiny';





#####################################################################
# Constructor and friends

# Create a new config
sub new {
	my $class = shift;

	# Create the object with default values
	my $self = $class->SUPER::new();
	$self->{_NEW} = 1;

	# Set some defaults for every Config object,
	# that will get overwritten.

	# There's a series of directories we looks for to find an obvious
	# case of a directory we SHOULD be using. The current two are ones most
	# likely to be found existing, and for which we would have write permissions.
	my @dirs = qw{/var/cache/httpd /var/cache/apache};
	foreach my $dir ( @dirs ) {
		next unless -e $dir;
		next unless -d $dir;
		next unless -r $dir;
		next unless -w $dir;

		# Oh goody, one that works!
		$self->{host}->{scratch} = $dir;
	}

	# Use a default if we couldn't get a prefered location
	$self->{host}->{scratch} ||= '/tmp';

	# Set some defaults for the cache filesystem and http locations
	$self->{host}->{staticfilesystem} = "$ENV{DOCUMENT_ROOT}/cvsmonitor";
	$self->{host}->{statichttp} = "$ENV{DOCUMENT_ROOT}/cvsmonitor";

	# Admin section password, by default, none
	$self->{security}->{adminpassword} = '';

	# The format to show the time in.
	# - "GMT". All values are in GMT
	# - "relative". All values are relative
	# - "smart". Recent times are relative, otherwise in GMT
	$self->{gui}->{timeformat} = 'smart';

	# The default number of changesets to show is 10
	$self->{gui}->{changesets} = 10;

	# Rebless as a CVSMonitor::Config object
	bless $self, $class;
}

# Load an existing config
sub load { return shift->read( @_ ) }
sub read {
	my $class = shift;

	# Check the file name
	my $file = shift;
	return undef unless -e $file;
	return undef unless -f $file;
	return undef unless -r $file;

	# Create the default configuration
	my $self = $class->new();

	# Get the on-disk config
	my $Config = $class->SUPER::read( $file )
		or return $class->_error( Config::Tiny->errstr );

	# Set the values from the file.
	# We use the setter methods so that if they have changed the config file,
	# the values are properly checked.
	foreach my $section ( keys %$Config ) {
		next if $section eq '_NEW';
		my $block = $Config->{$section};
		foreach my $key ( keys %$block ) {
			my $method = 'set' . ucfirst(lc($section)) . ucfirst(lc($key));
			next unless can( $self, $method );
			unless ( $self->$method( $block->{$key} ) ) {
				return $self->_error( "Bad value for '$section.$key' in config file '$file': " . $self->errstr );
			}
		}
	}

	delete $self->{_NEW};

	$self;
}

# Save a config ( wrap through to Config::Tiny->write )
sub save { shift->write( @_ ) }
sub write {
	my $self = shift;

	# Delete the _NEW tag on save
	delete $self->{_NEW};

	# Pass through
	$self->SUPER::write( @_ );
}

sub isNew { $_[0]->{_NEW} ? 1 : 0 }

# Get some values from the setup.conf file
sub importSetupValues {
	my $self = shift;

	# Just look in the default location
	my $file = './setup.conf';

	# Can we find the setup.conf file?
	return undef unless -e $file;
	return undef unless -f $file;
	return undef unless -r $file;

	# Get a Config::Tiny object for the file
	my $Config = Config::Tiny->read( $file ) or return undef;

	# Import the values
	unless ( $Config->{statichtml} ) {
		return $self->_error( 'Failed to find cache information in setup.conf. Have you run setup.pl yet?' );
	}
	$self->setHostStaticfilesystem( $Config->{statichtml}->{filesystem} )
		or $self->_error( 'Failed to set filesystem cache location' );
	$self->setHostStatichttp( $Config->{statichtml}->{http} )
		or $self->_error( 'Failed to set http cache location' );

	1;
}





#####################################################################
# Getters and Setters

# Get and set the time format
sub getGuiTimeformat { $_[0]->{gui}->{timeformat} }
sub setGuiTimeformat {
	my $self = shift;
	my $value = shift;
	return undef unless $value =~ /^(?:GMT|relative|smart)$/;

	$self->{gui}->{timeformat} = $value;

	1;
}

# Get and set the quantity of ChangeSets to show on pages
sub getGuiChangesets { $_[0]->{gui}->{changesets} }
sub setGuiChangesets {
	my $self = shift;
	my $value = int shift;
	if ( $value > 100 ) {
		$self->{gui}->{changesets} = 100;
	} elsif ( $value > 0 ) {
		$self->{gui}->{changesets} = $value;
	} else {
		delete $self->{gui}->{changesets};
	}

	1;
}

# Get and set the cache information
sub getHostStaticfilesystem { $_[0]->{host}->{staticfilesystem} }
sub setHostStaticfilesystem {
	my $self = shift;
	my $value = shift;

	# Does the location exist
	unless ( -e $value ) {
		return $self->_error( "Cache filesystem location '$value' does not exist" );
	}
	unless ( -d $value ) {
		return $self->_error( "Cache filesystem location '$value' is not a directory" );
	}
	unless ( -w $value ) {
		return $self->_error( "Cache filesystem location '$value' is not writable" );
	}

	$self->{host}->{staticfilesystem} = $value;

	1;
}

sub getHostStatichttp { $_[0]->{host}->{statichttp} }
sub setHostStatichttp {
	my $self = shift;
	my $value = shift;

	# Did they give us a value
	unless ( $value ) {
		return $self->_error( 'You did not enter a value for the cache http location' );
	}

	unless ( $value =~ m!^https?://! ) {
		return $self->_error( "The cache http location '$value' is not a http:// or https:// url" );
	}

	# Looks good
	$self->{host}->{statichttp} = $value;

	1;
}

# Get and set the working direcory.
# When setting the working directory, we make sure that we
# have write permission to the directory.
sub getHostScratch { $_[0]->{host}->{scratch} }
sub setHostScratch {
	my $self = shift;
	my $value = shift;

	# Does the directory they have specified exist?
	unless ( -e $value ) {
		return $self->_error( "The working directory does not exist" );
	}

	# Is it actually a directory
	unless ( -d $value ) {
		return $self->_error( "The working directory is a file, not a directory" );
	}

	# Can we read the directory
	unless ( -r $value ) { 
		return $self->_error( "Do not have read permissions to the working directory" );
	}

	# Can we write to the directory
	unless ( -w $value ) {
		return $self->_error( "Do not have write permissions to the working directory" );
	}

	# Looks good
	$self->{host}->{scratch} = $value;

	1;
}

# Get and set the password
sub getSecurityAdminpassword { $_[0]->{security}->{adminpassword} }
sub setSecurityAdminpassword {
	my $self = shift;
	my $value = shift;
	$self->{security}->{adminpassword} = $value;
	1;
}

1;

__END__

# Copyright (C) 2002-2004 Adam Kennedy
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Should you wish to utilise this software under a different licence, 
# please contact the author.
}
