#!/usr/bin/perl

# setup.pl 2 - More flexible program installation helper

BEGIN { require 5.005 }
use strict;
use UNIVERSAL 'isa';

# Make sure we are being called as a CGI script.
# Might write a console front end later though.
use vars qw{$VERSION};
BEGIN {
	$VERSION = 0.5;

	# Remove these for production
	#$ENV{HTTP_HOST} = 'ali.as';
	#$ENV{SCRIPT_NAME} = 'setup.pl';

	unless ( exists $ENV{HTTP_HOST} ) {
		die "This install script is a CGI application.\nPlease access it from a web browser\nHalted";
	}
}

# Declare all the globals
use vars qw{$config_file};
use vars qw{$dir $query $message};
use vars qw{$application %sections %cmds};
use vars qw{@section_order $test_file_name};

# Work out the directory the script is in,
# and make sure that directory is in the include path.
use FindBin (); 
BEGIN { 
	chdir( $dir = $FindBin::Bin );

	$config_file = './setup.conf';

	$message = '';

	# Initialise some simple variables
	$application = {
		name => 'the application'
		};

	# Set the section order
	@section_order = qw{legal perl filesystem html};

	# The name to use for the test file
	$test_file_name = '_setup_test.html';

	# Include our own module dirs
	@INC = ( './modules', @INC, './bundled', '.' );
}

#use lib '../../modules';  # Development only
#use lib '../cvsmonitor';  # Development only
use CGI;
BEGIN { $CGI::DEBUG = $CGI::DEBUG = 2 }

use setup::All;

# Define the error handler stuff REAL early
BEGIN {
	# Set the page source path
	setup::Page->setBasePath( './setup/html' );

	# Add all the subs we are going to 
	sub Error {
		# Show an error
		my $parser = new_parser();
		$parser->{error} = shift;
		$parser->{title} = 'Error';
		setup::Page->show( 'Error', $parser );
		exit();
	}

	sub new_parser {
		my $parser = setup::Parser->new;

		# Script callback
		$parser->{callback} = $ENV{SCRIPT_NAME};

		# Timestamp and copyright for the bottom bar
		$parser->{timestamp} = scalar localtime time;
		my @t = localtime time;
		$parser->{copyright} = "Ali.as Setup $VERSION &copy; Adam Kennedy 2001 - " 
			. ($t[5] + 1900);

		# Add the web user and directory to the default parser
		$parser->{web_user} = `whoami`;
		chomp $parser->{web_user};

		# Add the directory we are running from
		$parser->{web_directory} = $dir;
		$parser->{web_filename} = $ENV{SCRIPT_FILENAME};

		$parser;
	}
}

# Create some more general globals
BEGIN {
	# Set the callback name
	unless ( $ENV{SCRIPT_NAME} ) {
		Error( "Failed to get script name for callback" );
	}

	# Some variables
	%sections = (
		legal      => 'Legal Requirements',
		perl       => 'Perl Requirements',
		filesystem => 'File System Requirements',
		html       => 'HTML Installation',
		);
	%cmds = (
		legal      => 'viewLegal',
		perl       => 'viewPerl',
		filesystem => 'viewFileSystem',
		html       => 'viewHTML',
		);
}

# Get the query
$query = CGI->new() or Error( "Failed to get CGI query" );

# Check the config file
unless ( -e $config_file ) {
	Error( 'Failed to find setup.conf file' );
}
unless ( -r $config_file ) {
	Error( 'setup.conf file exists, but I do not have permissions to read it'
		. 'Please make the setup.conf file readable by me' );
}

# Try to load the config file
my $Config = Config::Tiny->read( $config_file );
Error( 'Unable to load setup.conf file' ) unless $Config;

# Set the application information
if ( $Config->{application} ) {
	$application = { %{ $Config->{application} } };
	$application->{name} = 'the application' unless $application->{name};
}

# Is the config file writable
unless ( -w $config_file ) {
	ErrorConfig();
}

# Create the command map
my $cmdmap = {
	viewIntroduction => \&viewIntroduction,

	viewLegal        => \&viewLegal,
	viewPerl         => \&viewPerl,
	viewFileSystem   => \&viewFileSystem,

	viewHTML         => \&viewHTML,
	testHTML         => \&cmdTestHTML,
	confirmHTML      => \&cmdConfirmHTML,
	installHTML      => \&cmdInstallHTML,

	viewConclusion   => \&viewConclusion,

	'reset'          => \&cmdReset,
	};

# Get the command
my $cmd = $query->param( 'cmd' ) || 'viewIntroduction';

# Dispatch the command
unless ( defined $cmdmap->{$cmd} ) {
	Error( "Invalid script action '$cmd'" );
}
&{$cmdmap->{$cmd}};
exit();





#####################################################################
# Views

sub viewIntroduction {
	# Create the parser
	my $parser = new_parser();
	$parser->{title} = "Introduction";

	# Show the page
	show_page( 'Introduction', $parser );
}

sub viewLegal {
	# Create the parser
	my $parser = new_parser();
	$parser->{title} = $sections{legal};

	# Show the page
	show_page( 'Legal', $parser );
}

sub viewPerl {
	# Create the parser
	my $parser = new_parser();
	$parser->{title} = $sections{perl};

	# Get the section's hash
	my $modules = {%{$Config->{perl}}};

	# Is there a problem?
	my $problem = 0;

	# Check the Perl version
	if ( $modules->{version} ) {
		if ( setup::Tests->perlVersion( $modules->{version} ) ) {
			$parser->{version} = qq~
			  <tr><td colspan=2 nowrap bgcolor='#99FF99'>Perl version is new enough 
			    ( Found: <b>$]</b> | Needed: <b>$modules->{version}</b> )</td></tr>
			  ~;
		} else {
			$parser->{version} = qq~
			  <tr>
			    <td colspan=2 nowrap bgcolor='#FF9999'>Perl version is too old 
			    ( Found: <b>$]</b> | Needed: <b>$modules->{version}</b> )</td>
			  </tr>
			  ~;
			$problem++;
		}
		delete $modules->{version};
	} else {
		$parser->{version} = qq~
		  <tr><td colspan=2 nowrap bgcolor='#99FF99'>No Perl version requirement</td></tr>
		  ~;
	}

	# Is there a bundle available.
	# We will use this later if problems arise.
	if ( $modules->{bundle} ) {
		my ($bundle_name, $bundle_version) = split /\s+/, $modules->{bundle};
		my $installed = setup::Tests->perlModule( $bundle_name );
		if ( $bundle_version and $installed < $bundle_version ) {
			$parser->{overall} .= qq~
			  <tr>
			    <td colspan=2 bgcolor='#FFFFFF'><p>A newer version <b><font color="#000099">$bundle_version</font></b>
			   	  of <b><font color="#000099">$bundle_name</font></b>
			      is available to assist you with module installation.</p>
			      <p>If you have root access, you can type</p>
			      <pre>perl -MCPAN -e 'install $bundle_name'</pre>
			      <p>from the console to install the required CPAN modules.</p>
			      <p><b>If at all possible</b>, this is the recommended way to install the modules</p>
			    </td>
			  </tr>
			  ~;

		} elsif ( ! $installed ) {
			$parser->{overall} .= qq~
			  <tr>
			    <td colspan=2 bgcolor='#FFFFFF'><p>The bundle <b><font color="#000099">$bundle_name</font></b>
			      is available to assist you with module installation.</p>
			      <p>If you have root access, you can type</p>
			      <pre>perl -MCPAN -e 'install $bundle_name'</pre>
			      <p>from the console to install the required CPAN modules.</p>
			      <p><b>If at all possible</b>, this is the recommended way to install the modules</p>
			    </td>
			  </tr>
			  ~;
		}
		delete $modules->{bundle};
	}		  		  

	# Go through and check each one
	my @cpan_required_modules = ();
	my @cpan_optional_modules = ();
	my @private_modules = ();

	foreach my $module_name ( sort keys %$modules ) {
		# Get the module information
		my ($module_type, $module_version) = split /\s+/, $modules->{$module_name};
		my $installed = setup::Tests->perlModule( $module_name );

		# Create the 'Found' string
		my $words;
		if ( $installed ) {
			$words = 'Found' . ($installed != -1
				? " $installed" : '');
		} else {
			$words = 'Not Found';
		}

		# Is it good enough?
		my $colour;
		if ( $module_version and $installed < $module_version ) {
			# Not new enough
			$colour = '#FF9999'; # Red
			$problem++;

		} elsif ( $installed ) {
			# Found and new enough
			$colour = '#99FF99'; # Green

		} else {
			# Not installed
			$colour = '#FF9999'; # Red
			$problem++;
		}

		# Create the row
		$module_version = ">= $module_version" if $module_version;
		my $row = qq~
		  <tr>
		    <td nowrap bgcolor='$colour'>$module_name $module_version</td>
		    <td nowrap bgcolor='$colour'>$words</td>
		  </tr>
		  ~;

		# Add the row to the appropriate section
		if ( $module_type eq 'required' ) {
			push @cpan_required_modules, $row;
		} elsif ( $module_type eq 'optional' ) {
			push @cpan_optional_modules, $row;
		} elsif ( $module_type eq 'private' ) {
			push @private_modules, $row;
		} else {
			# What the hell is this?
			Error( "Unknown perl module check type '$module_type' for module '$module_name'" );
		}
	}

	# Set the parser values
	if ( scalar @cpan_required_modules ) {
		$parser->{required} = join '', @cpan_required_modules;
	} else {
		$parser->{required} = qq~<tr>
		    <td colspan=2 nowrap bgcolor='#99FF99'>No Required Modules</td>
		  </tr>~;
	}
	if ( scalar @cpan_optional_modules ) {
		$parser->{optional} = join '', @cpan_optional_modules;
	} else {
		$parser->{optional} = qq~<tr>
		    <td colspan=2 nowrap bgcolor='#99FF99'>No Optional Modules</td>
		  </tr>~;
	}
	if ( scalar @private_modules ) {
		$parser->{private} = join '', @private_modules;
	} else {
		$parser->{private} = qq~<tr>
		    <td colspan=2 nowrap bgcolor='#99FF99'>No Private Modules</td>
		  </tr>~;
	}

	# Add the overall result, and the link to the next section
	if ( $problem ) {
		$parser->{overall} = qq~
		  <tr> 
		    <td nowrap bgcolor="#FF9999">Insufficient</td>
		    <td bgcolor="#FFFFFF">Resolve the problems, and reload the page</td>
		  </tr>
		  ~;
	} else {
		# Tag the section as completed, and write to disk
		$Config->{status}->{perl} = 'completed';
		$Config->write( $config_file );

		my $next_link = fragmentNextLink();
		$parser->{overall} = qq~
		  <tr> 
		    <td nowrap bgcolor="#99FF99">Satisfactory</td>
		    <td bgcolor="#FFFFFF">$next_link</td>
		  </tr>
		  ~;

	}

	# Show the page
	show_page( 'Perl', $parser );
}

sub viewFileSystem {
	# Create the parser
	my $parser = new_parser();
	$parser->{title} = $sections{filesystem};

	# Get the filesystem block
	my $filesystem = {%{$Config->{filesystem}}};
	Error( "No filesystem section found" ) unless $filesystem;

	# Add the results for the 'execute script' text
	my @tests = ();
	my @output = ();

	# Iterate over the tests
	my $problem = 0;
	foreach my $target ( keys %$filesystem ) {
		my $command = $filesystem->{$target};
		my $english = {
			'exists'     => 'find',
			canread      => 'read',
			canwrite     => 'write',
			canreadwrite => 'read and write',
			canexecute   => 'execute',
			installed    => 'find installed program',
			}->{$command};
		my $method = {
			'exists'       => 'filesystemExists',
			'canread'      => 'filesystemCanRead',
			'canwrite'     => 'filesystemCanWrite',
			'canreadwrite' => 'filesystemCanReadWrite',
			'canexecute'   => 'filesystemCanExecute',
			'installed'    => 'programInstalled',
			}->{$command};

		# Do the test
		my ($colour, $words);
		if ( setup::Tests->$method( $target ) ) {
			$colour = '#99FF99';
			$words = 'Passed';
		} else {
			$colour = '#FF9999';
			$words = 'Failed';
			$problem++;
		}

		# If the command is a filesystem one, absolute it
		if ( $command =~ /^(?:exists|canread|canwrite|canreadwrite|canexecute)$/ ) {
			$english = "User '$parser->{web_user}' can " . $english;
			$target =~ s!^\.\/!!;
			$target = "$FindBin::Bin/$target";
		}

		push @output, qq~<tr>
		    <td nowrap bgcolor="$colour">$english $target</td>
		    <td nowrap bgcolor="$colour" align='center'>$words</td>
		  </tr>~;
	}
	$parser->{required} = join '', @output;

	# Add the overall result, and the link to the next section
	if ( $problem ) {
		$parser->{overall} = qq~
		  <tr> 
		    <td nowrap bgcolor="#FF9999">Insufficient</td>
		    <td bgcolor="#FFFFFF">Resolve the problems, and reload the page</td>
		  </tr>
		  ~;
	} else {
		$Config->{status}->{filesystem} = 'completed';
		$Config->write( $config_file );

		my $next_link = fragmentNextLink();
		$parser->{overall} = qq~
		  <tr> 
		    <td nowrap bgcolor="#99FF99">Satisfactory</td>
		    <td bgcolor="#FFFFFF">$next_link</td>
		  </tr>
		  ~;
	}

	# Show the page
	show_page( 'FileSystem', $parser );
}

sub viewHTML {
	# Create the parser
	my $parser = new_parser();
	$parser->{title} = $sections{html};

	if ( $Config->{statichtml} ) {
		return viewHTMLInstall();
	}

	# Set the values for the text fields
	$parser->{root_filesystem} = $query->param( 'root_filesystem' );
	unless ( $parser->{root_filesystem} ) {
		my $root = $ENV{DOCUMENT_ROOT};
		$root =~ s!/+$!!;
		$parser->{root_filesystem} = "$root/cvsmonitor";
	}
	$parser->{root_http} = $query->param( 'root_http' )
		|| "http://$ENV{SERVER_NAME}/cvsmonitor";

	if ( length $message ) {
		$message = qq~
       		 <table border="0" cellspacing="0" cellpadding="5">
      		    <tr>
            		<td><font color="#990000"><b>$message</b></font></td>
      		    </tr>
      		  </table>
		~;
	}

	# Show the page
	show_page( 'HTML', $parser );
}

# Test the HTML location information
sub cmdTestHTML {
	# Did they give us the filesystem location
	my $root = {
		fs => $query->param( 'root_filesystem' ),
		web => $query->param( 'root_http' ),
		};
	unless ( $root->{fs} ) {
		$message = 'You did not enter a value for the filesystem';
		return viewHTML();
	}
	unless ( $root->{web} ) {
		$message = 'You did not enter a value for the web location';
		return viewHTML();
	}

	# Trim trailing slashes
	$root->{fs} =~ s/\/$//;
	$root->{web} =~ s/\/$//;

	# Does the filesystem directory they specify exist?
	if ( -e $root->{fs} ) {
		unless ( -d $root->{fs} ) {
			$message = "'$root->{fs}' exists, but it is a file";
			return viewHTML();
		}
	} else {
			# Could we create it?
			unless ( File::Flat->canWrite( $root->{fs} ) ) {
				$message = "The directory '$root->{fs}' does not exist, and I don't have the enough rights to create it.<br>"
					. "You probably want to do something like the following:<br>"
					. "mkdir $root->{fs}<br>"
					. "chmod a+rwx $root->{fs}";
				return viewHTML();
			}
	}

	# Where will this file go
	my $file = {
		fs => "$root->{fs}/$test_file_name",
		web => "$root->{web}/$test_file_name",
		};

	# Can we write to the filesystem location
	unless ( File::Flat->canWrite( $file->{fs} ) ) {
		$message = "Insufficient permissions to create test file '$file->{fs}'.<br>"
			. "You probably want to do something like the following:<br>"
			. "chmod a+rwx $root->{fs}";
		return viewHTML();
	}

	# Create the test file
	my $page = generateTestPage( $root );
	unless ( $page ) {
		Error( "Failed to generate test page" );
	}

	# Save the test page
	unless ( $page->save( $file->{fs} ) ) {
		Error( "Error saving test page" );
	}

	# Right, now show them the about to test page
	my $parser = new_parser();
	$parser->{redirect} = $file->{web};
	show_page( 'TestStarting', $parser );
}

sub generateTestPage {
	my $root = shift;

	my $parser = new_parser();
	$parser->{root_filesystem} = $root->{fs};
	$parser->{root_http} = $root->{web};
	$parser->{file_filesystem} = "$root->{fs}/$test_file_name";
	$parser->{file_http} = "$root->{web}/$test_file_name";

	# Create the page
	my $Page = setup::Page->new( 'Test', $parser ) or return undef;

	# Return it
	return $Page;
}

sub cmdConfirmHTML {
	# Make sure we haven't lost the values
	unless ( $query->param( 'root_filesystem' ) ) {
		Error( 'Lost the html filesystem root value' );
	}
	unless ( $query->param( 'root_http' ) ) {
		Error( 'Lost the html web root value' );
	}

	# Set the information
	$Config->{statichtml}->{filesystem} = $query->param( 'root_filesystem' );
	$Config->{statichtml}->{http} = $query->param( 'root_http' );
	$Config->write( $config_file );

	# Move on to the installation page
	return viewHTMLInstall();
}

sub viewHTMLInstall {
	# Make sure we know the static html stuff
	unless ( $Config->{statichtml} ) {
		Error( "Can't do installations when we don't know where to install to" );
	}

	# Go through each of the installation sections
	my $resources = $Config->{html}
		or Error( 'Failed to locate anything to install' );

	my $problem = 0;
	my @rows = ();
	foreach ( sort keys %$resources ) {
		my $resource = $_;
		my $source = $resources->{$_};

		# Does the resource exist?
		my $color = '';
		my $installed = '';
		my $command = '&nbsp;';
		my $dir = "$Config->{statichtml}->{filesystem}/$resource";
		if ( -d $dir ) {
			# Yes, good
			$installed = 'Yes';
			$color = '#99FF99';
			if ( $source ne 'create' ) {
				$command = fragmentLink( 'Force Re-Install', {
					cmd => 'installHTML',
					resource => $resource,
					} );
			}

		} elsif ( -f $dir ) {
			# There's a file in the way
			$color = '#FF9999';
			$installed = 'No. File blocks installation';
			$command = "Will not risk overwriting file...<br>"
				. "Check and remove by hand before continuing";
			$problem++;

		} else {
			# Nope
			$color = '#FF9999';
			$installed = 'No';
			$command = fragmentLink( ($source eq 'create')
				? 'Create Directory'
				: 'Install Directory', 
				{ cmd => 'installHTML', resource => $resource } 
				);
			$problem++;

		}

		# Add to the rows
		push @rows, qq~
		  <tr bgcolor="$color">
             <td>$resource</td>
             <td>$source</td>
             <td>$installed</td>
             <td bgcolor='#FFFFFF'>$command</td>
          </tr>
          ~;
	}

	# Add the rows to the page
	my $parser = new_parser();
	$parser->{title} = $sections{html};
	$parser->{rows} = join "\n", @rows;

	# Add the overall result
	if ( $problem ) {
		$parser->{overall} = qq~
		  <tr> 
		    <td nowrap bgcolor="#FF9999">Insufficient</td>
		    <td bgcolor="#FFFFFF">Resolve the problems, and reload the page</td>
		  </tr>
		  ~;
	} else {
		$Config->{status}->{html} = 'completed';
		$Config->write( $config_file );

		my $next_link = fragmentNextLink();
		$parser->{overall} = qq~
		  <tr> 
		    <td nowrap bgcolor="#99FF99">Satisfactory</td>
		    <td bgcolor="#FFFFFF">$next_link</td>
		  </tr>
		  ~;
	}

	# Show the completed page
	show_page( 'HTMLInstall', $parser );
}

sub cmdInstallHTML {
	# Does the resourse exist?
	my $resource = $query->param( 'resource' );
	unless ( $resource ) {
		Error( "You did not specify a resource" );
	}
	unless ( defined $Config->{html}->{$resource} ) {
		Error( "No such resource '$resource' to install" );
	}

	# Establish the destination
	unless ( $Config->{statichtml} ) {
		Error( 'Lost the cache information' );
	}
	my $target = "$Config->{statichtml}->{filesystem}/$resource";

	# Is it a path, or a create
	my $source = $Config->{html}->{$resource};
	if ( $source eq 'create' ) {
		if ( -d $target ) {
			$message = "Directory $target already exists";
		} elsif ( -f $target ) {
			$message = "Directory creation blocked by file $target";
		} else {
			# Try to create the directory
			my $rv = File::Flat->makeDirectory( $target );
			$message = $rv
				? "Created directory $target"
				: ("Failed to create directory $target" . File::Flat->errstr) ;
		}
	} else {
		# If target is a file, die
		if ( -f $target ) {
			$message = "Directory installation blocked by file $target";
		} elsif ( -d $target ) {
			# Delete the existing directory
			my $rv = File::Flat->remove( $target );
			unless ( $rv ) {
				$message = "Failed to remove existing directory $target";
			}
		}

		# Install the new directory
		my $rv = File::Flat->copy( $source, $target );
		$message = $rv
			? "Installed $resource from $source to $target"
			: "Failed to install $resource from $source to $target";
	}

	# Return the the Install page
	viewHTMLInstall();
}

sub viewConclusion {
	# Create the parser
	my $parser = new_parser();
	$parser->{title} = 'Conclusion';
	$parser->{application_start} = $Config->{application}->{start};

	# Show the page
	show_page( 'Conclusion', $parser );
}

sub cmdReset {
	$Config->{status} = {};
	$Config->{statichtml} = {};
	$Config->write( $config_file );

	# Now go on to the introduction page
	return viewIntroduction();
}




#####################################################################
# Fragment generators

sub fragmentNextLink {
	foreach ( @section_order ) {
		next unless $Config->{$_};
		unless ( $Config->{status}->{$_} eq 'completed' ) {
			return fragmentLink( "Continue to $sections{$_}", { cmd => $cmds{$_} } );
		}
	}
	fragmentLink( 'Finish Installation', { cmd => 'viewConclusion'} );
}

# Create a generic callback link
sub fragmentLink {
	my $text = shift;
	my $args = shift || {};
	return 'Error' unless isa( $args, 'HASH' );

	# Crap to hopefully avoid warnings
	foreach ( keys %$args ) {
		$args->{$_} = '' unless defined $args->{$_};
	}

	my $arg_string = join( '&', map { "$_=$args->{$_}" } keys %$args );
	return length($arg_string)
		? "<a href='$ENV{SCRIPT_NAME}?$arg_string'>$text</a>"
		: "<a href='$ENV{SCRIPT_NAME}'>$text</a>";
}

sub fragmentStatusBox {
	# Generate a status box
	my $colour = {};

	# Get the colours for each section
	foreach ( @section_order ) {
		$colour->{$_} = $Config->{$_}
			? ($Config->{status}->{$_} eq 'completed')
				? '#009900' : '#990000'
			: '#999999';
	}

	return qq~
	<table border="0" cellspacing="0" cellpadding="1" align="right">
	  <tr>
	    <td bgcolor="#999999"> 
	      <table border="0" cellspacing="1" cellpadding="3">
	        <tr>
	          <td bgcolor="#CCCCCC" nowrap><b>Install Status</b></td>
	        </tr>
	        <tr>
	          <td bgcolor="#FFFFFF" nowrap> 
	              - <b>Introduction</b><br>
	              - <b><font color="$colour->{legal}">Legal</font></b><br>
                      - <b><font color="$colour->{perl}">Perl</font></b><br>
	              - <b><font color="$colour->{filesystem}">FileSystem</font></b><br>
	              - <b><font color="$colour->{html}">HTML</font></b>
	            </td>
	        </tr>
	      </table>
	    </td>
	  </tr>
	</table>
	~
}





#####################################################################
# Utility functions

# Add the toolbar and show the page
sub show_page {
	my $page = shift;
	my $parser = shift || new_parser();

	# Add the message
	$parser->{message} = $message if $message;

	# Add the application stuff
	$parser->{application_name} = $application->{name};
	$parser->{application_codename} = $application->{codename};
	$parser->{application_start} = $application->{start};

	# Add the status box
	$parser->{status} = fragmentStatusBox();

	# Add the next link, if they want it
	$parser->{next_link} = fragmentNextLink();

	# Show the templated page
	return if setup::Page->show( $page, $parser );
	Error( "Error trying to show page<br>" . setup::Error->errstrHTML );
}

# Tell them to make the config file writable
sub ErrorConfig {
	my $parser = new_parser();
	$parser->{title} = 'Before we begin...';

	# Add the application stuff
	$parser->{application_name} = $application->{name};
	$parser->{application_codename} = $application->{codename};
	$parser->{application_start} = $application->{start};

	setup::Page->show( 'ErrorConfig', $parser );
	exit();
}

1;
