package setup::Page;

# The setup::Page module handles the load, parsing, and display
# of HTML pages

use strict;
use UNIVERSAL 'isa';
use CGI;
use File::Flat;
use vars qw{@ISA};
BEGIN { @ISA = 'setup::Error' }




#####################################################################
# Constructor and main shortcut functions

# Create a new page
sub new { 
	my $class = shift;
	my $name = shift or return undef;
	isa( $_[0], 'setup::PageFactory' )
		? shift()->page($name, shift)
		: setup::PageFactory->page($name, shift);
}
		
# Make a new page and display it
sub show {
	my $Page = shift->new( @_ ) or return undef;
	$Page->display ? 1 : undef;
}

# Make a new page and save it
sub freeze {
	my $file = $_[3] or return undef;
	my $Page = shift->new( @_ ) or return undef;
	$Page->save( $file ) ? 1 : undef;
}





#####################################################################
# Changing Page Properties

# Read-Only properties
sub Name     { $_[0]->{name} }
sub FileName { $_[0]->{filename} }

# Get and set the parser for the Page
sub getParser { $_[0]->{parser} }
sub setParser {
	my $self = shift;
	my $parser = shift;

	# Make the parser an object if it isn't
	if ( isa( $parser, 'setup::Parser' ) ) {
		$self->{parser} = $parser;
		return 1;

	} elsif ( isa( $parser, 'HASH' ) ) {
		$self->{parser} = setup::Parser->new( $parser ) || return undef;
		return 1;

	} else {
		return $self->andError( "Invalid parser argument" );
	}
}

# Get and set the source content
sub getSource {
	my $self = shift;

	# Load the source if needed
	unless ( defined $self->{source} ) {
		my $source = File::Flat->slurp( $self->{filename} );
		return $self->andError( "Error trying to load page '$self->{name}'" ) unless $source;
		$self->{source} = $$source;
	}

	$self->{source};
}
sub setSource { $_[0]->{source} = $_[1]; 1; }

# Get and set the parsed content
sub getParsed {
	my $self = shift;

	# Shortcut if there is no parser
	return $self->getSource unless $self->{parser};
	
	# Get the source
	my $content = $self->getSource;
	return undef unless defined $content;
	my $parsed = $self->{parser}->parse( $content );
	return $self->andError( "Error parsing page" ) unless defined $parsed;

	# Hand off to our Factory to filter
	$self->filter( $parsed ) 
		|| $self->andError( "Error filtering page" );
}

# Add a cookie to this particular page
sub setCookie {
	my $either = shift;
	my $cookie = shift;
	return undef unless isa( $cookie, 'CGI::Cookie' );
	if ( ref $either ) {
		# Add the cookie to this page
		if ( $either->{cookies} ) {
			push @{$either->{cookies}}, $cookie;
		} else {
			$either->{cookies} = [ $cookie ];
		}
	} else {
		# Pass through if called statically
		setup::PageFactory->setCookie( $cookie );
	}
	1;
}	





	
	
#####################################################################
# Outputing Page Content

# Display the page
sub display {
	# Make sure we can get the page, then send
	my $content = $_[0]->getParsed or return undef;
	my $header = $_[0]->header or return undef;
	print STDOUT $header . $content;
	1;
}

# Generate the page header
sub header {
	my $either = shift or return undef;
	if ( ref $either and $either->{cookies} ) {
		# Detach the cookies, and
		# generate a header with them
		my $cookies = $either->{cookies};
		$either->{cookies} = undef;
		return CGI::header( '-cookie', $cookies );
	} else {
		# Send a standard header
		return CGI::header();
	}
}

# Instead of displaying a page, 
# save it to the file system
sub save {
	my $self = shift;
	my $filename = shift;

	# Make sure we have the parsed page
	my $content = $self->getParsed;
	return undef unless defined $content;

	# Reformat the page for our filesystem, and save
	$content =~ s/(\015\012|\015|\012)/\n/g;
	File::Flat->write( $filename, \$content ) ? 1 : undef;
}

# Filter the parsed page
sub filter {
	my $self = shift;
	my $content = shift;
	return $content unless $self->{filter};

	# Pass the content to each of the filters
	foreach my $filter ( @{ $self->{filter} } ) {
		$content = &$filter( $self, $content );
		return undef unless defined $content;
	}

	$content;
}





#####################################################################
# Compatibility Methods
#
# Pass these through to the default PageFactory

sub setBasePath { setup::PageFactory->setBasePath( $_[1] ) }

	
	




#####################################################################
# Utility functions

# Scan the page for matching parser tags
sub scan {
	my $self = shift;
	unless ( ref $self ) {
		# Make a new object with the name
		$self = $self->new( shift ) or return undef;
	}
	my $tagroot = shift;

	# Make sure we have the source
	$self->getSource or return undef;
	
	# Scan the source and return the results
	my $results = setup::Parser->scan( $self->{source}, $tagroot );
	$results ? $results : undef;
}

# Write an error message using nothing outside this function
# and nothing that can fail
sub safeError {
	shift;
	my $message = join "<br>", @_;
	print "Content-type: text/html\n\n";
	print qq~<html><head><title>Error</title></head>
		<body><h1>Error</h1>
		<p><font face="Verdana, Arial, Helvitica" size="2" color="#990000"><b>$message</b></font></p>~
		. setup::Error->calltraceHTML()
		. qq~</body></html>~;
}

1;
